# *- coding: utf-8 -*-
""" decorators

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import inspect
import sys
import functools
import weakref
import six
import wrapt
from wrapt import decorator

from pprint import pformat
from pyrsistent import pmap
from jsonschema.exceptions import ValidationError

from .exceptions import InvalidValue

# about decorators and why using wrapt
# https://hynek.me/articles/decorators/


def take_arrays(narg1=0, narg2=-1, flatten=False):
    """
    Decorator for functions/methods to take a list input and retrieve a list as
    a result

    :param narg1: index of argument to use as criteria
    :param narg2: index of an optional second argument (can be a list of same
                  size, or a scalar)
    :param flatten: flatten results
    :rtype: returns a list if the input argument is a list
    """

    def to_decorate(wrapped):
        sig = inspect.getargspec(wrapped)

        @wrapt.decorator
        def wrapper(method, instance, args, kwargs):
            # case first argument is an array
            if type(args[narg1]) in (list, set, tuple):
                ret = []
                new_args = list(args)
                # only first argument is an array
                if (narg2 == -1 or not type(args[narg2]) in (list, set, tuple)
                        or len(args[narg2]) == 1):
                    for v in args[narg1]:
                        new_args[narg1] = v
                        ret.append(wrapped(*new_args, **kwargs))
                # both are list
                else:
                    # same size - call with zipped values
                    if len(args[narg1]) == len(args[narg2]):
                        for v1, v2 in zip(args[narg1], args[narg2]):
                            new_args[narg1] = v1
                            new_args[narg2] = v2
                            ret.append(wrapped(*new_args, **kwargs))
                    # different size
                    else:
                        raise InvalidValue(
                            "arguments %s and %s must be of same size" %
                            (narg1, narg2))
            else:
                # if the first one is not a string, ignore and call func
                return wrapped(*args, **kwargs)
            if flatten:
                return [item for sublist in ret for item in sublist]
            return ret

        decorated = wrapper(wrapped)

        doc = [wrapped.__doc__.strip()] if wrapped.__doc__ else []
        doc.append(
            "Argument '%s' can take an array, " % (sig.args[narg1]) +
            "and the function will perfom the operation on each element " +
            "and return the results as a list.")
        if narg2 != -1:
            doc.append(
                "Arguments '%s' and '%s'" %
                (sig.args[narg1], sig.args[narg2]) +
                " can take arrays of same size, and the function will perfom " +
                "the operation in parallel and return the results as a list.")
        if flatten:
            doc.append("The resulting list of lists if flattened.")

        wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", "\n".join(doc))

        return decorated

    return to_decorate


logger = logging.getLogger(__name__)


def assert_arg(arg, typ, **schema):
    """
    Decorator to add a schema to validate a given argument against a json-schema.
    If the decorated function has a keyword argument `assert_args`, it is used as
    a flag to enable/disable argument validation/conversion.

    :param arg: argument to convert/validate, can be position (start 0) or name
    :type type: Type subclass
    :param schema: json-schema for the type
    """
    validator = typ(**schema)

    def to_decorate(wrapped):
        # find argument in signature
        arg_i = None
        arg_s = arg
        try:
            sig = inspect.getargspec(wrapped)
            if type(arg) is int:
                arg_i = arg
                arg_s = sig.args[arg]
            elif arg in sig.args:
                arg_i = sig.args.index(arg)
                arg_s = arg
        except Exception as er:
            logger.error(er)
            raise
            # if no inspection possible (if creating type metaprogramming)

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):

            # make assert optional with kwargs
            assert_args = kwargs.pop('assert_args', True)
            if not assert_args:
                return wrapped(*args, **kwargs)

            arg_i2 = arg_i if not instance and arg_i is not None else arg_i - 1
            args = list(args)
            try:
                if arg_s in kwargs:
                    kwargs[arg_s] = validator(kwargs[arg_s], validate=True)
                elif type(arg_i2) is int and arg_i2 < len(args):
                    args[arg_i2] = validator(args[arg_i2], check=True, validate=True)
                else:
                    # must be a default value, assume it s correct!
                    pass
                    #raise Exception("error with argument definition (%s,%i)"%(arg_s, arg_i2))
            except (ValidationError, InvalidValue) as er:
                if arg_s in kwargs:
                    raise ValidationError(
                        "%s=%r is not valid. %s" % (arg_s, kwargs[arg_s], er))
                elif type(arg_i) is int and arg_i2 < len(args):
                    raise ValidationError(
                        "%s=%r is not valid. %s" % (arg_s, args[arg_i2], er))
            return wrapped(*args, **kwargs)

        decorated = wrapper(wrapped)

        try:
            doc = (
                (wrapped.__doc__ or "").strip() + "\n\nArgument '%s' is " % arg +
                "automatically type converted and validated against %s."
                % pformat(validator))

            wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", doc)
        except Exception as er:
            logger.error(er)
            raise

        return decorated

    return to_decorate


def _format_call_msg(funcname, args, kwargs):
    """
    Format a call message
    """
    return "%s(%s)" % (
        funcname,
        ", ".join(["%r" % a for a in args] +
                  ["%s=%r" % (a, v) for a, v in kwargs.items()]),
    )


@wrapt.decorator
def log_exceptions(method, instance, args, kwargs):
    """
    log calls to a function/method with its arguments and the exceptions
    that might be raised.

    SHOULDN T BE USED FOR __init__, use log_init instead
    """
    # special case happening for getters/setters
    if instance is None:
        instance = args[0]
    try:
        if hasattr(instance, "_logger"):
            instance._logger.debug(
                "CALL %s",
                _format_call_msg(
                    "%r.%s" % (instance, getattr(method, '__name__', 'unknown')),
                    args, kwargs))
        return method(*args, **kwargs)
    except Exception as er:
        etype, value, trace = sys.exc_info()
        if hasattr(instance, "_logger"):
            instance._logger.error(
                "CALL %s",
                _format_call_msg(
                    "%r.%s" % (instance, getattr(method, '__name__', 'unknown')),
                    args, kwargs) +
                "\n\tERROR %s: %s" % (etype.__name__, value), exc_info=True)
        try:
            six.reraise(etype, value, trace)
        finally:
            del trace  # to avoid circular refs


@wrapt.decorator
def log_init(init, instance, args, kwargs):
    """
    log init of instance and possible exceptions
    """
    instance._logger.info(
        "%s",
        _format_call_msg("INIT <%s>.__init__" % instance.__class__.__name__,
                         args, kwargs))
    try:
        return init(*args, **kwargs)
    except Exception as er:
        etype, value, trace = sys.exc_info()
        instance._logger.error(
            "CALL %s\n\tERROR: %s: %s",
            _format_call_msg(
                "INIT <%s>.__init__" % instance.__class__.__name__, args,
                kwargs), etype.__name__, value)
        try:
            six.reraise(etype, value, trace)
        finally:
            del trace  # to avoid circular refs


def assert_prop(*args2check):
    """
    Assert some properties/attributes are defined in instance

    :param args2check: list of property/attribute to check as defined
    """

    def to_decorate(wrapped):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            for prop in args2check:
                if getattr(instance, prop, None) is None:
                    raise AttributeError("%s is not defined." % prop)
            return wrapped(*args, **kwargs)

        decorated = wrapper(wrapped)

        doc = (wrapped.__doc__ or "").strip()
        for prop in args2check:
            doc += "\nProperty %s is asserted to be defined before call." % prop
        wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", doc)

        return decorated

    return to_decorate


def depend_on_prop(*prop_names):
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        return wrapped(*args, **kwargs)
    return wrapper


class ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


class memoized_property(object):
    """A read-only @property that is only evaluated once."""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result

    def _reset(self, obj):
        memoized_property.reset(obj, self.__name__)

    @classmethod
    def reset(cls, obj, name):
        obj.__dict__.pop(name, None)


def memoized_method(*lru_args, **lru_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)
            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)
            # can t use setattr as it conflicts with __setattr__ of protocol base
            object.__setattr__(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)
        return wrapped_func
    return decorator



# -*- coding: utf-8 -*-
"""
Unit tests for decorators

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import unicode_literals

import logging
from builtins import object

import pytest
from python_jsonschema_objects.validators import ValidationError

from ngoschema.types import Integer
from ngoschema.decorators import assert_arg
from ngoschema.decorators import assert_prop
from ngoschema.decorators import log_exceptions
from ngoschema.decorators import log_init
from ngoschema.decorators import take_arrays
from ngoschema.exceptions import InvalidValue


@take_arrays()
def time2(d):
    """multiply by 2"""
    return 2 * d


@take_arrays(0, 1)
def add(a, b):
    """add 2 components
    :param a: 1st
    :param b: 2nd
    """
    return a + b


def test_decorators():
    assert time2(1) == 2
    assert time2([1, 2]) == [2, 4]
    assert add(1, 2) == 3
    assert add([1, 2], [2, 3]) == [3, 5]
    assert add([1, 2], 3) == [4, 5]
    with pytest.raises(Exception):
        add(3, [1, 2])
    with pytest.raises(InvalidValue):
        add([1, 2], [1, 2, 3])

    class MyException(Exception):
        pass

    class A(object):
        logger = logging.getLogger("TEST")

        @log_init
        def __init__(self):
            pass

        def __repr__(self):  # similar to the one in ClassWithSchema
            ret = "<%s" % self.__class__.__name__
            return ret + ">"

        def __str__(self):
            return self.__repr__()

        @log_exceptions
        def raise_exc(self):
            raise MyException("YO")

        @log_exceptions
        @assert_arg(1, Integer)
        def foo(self, integer):
            return 1 + integer

        @log_exceptions
        @assert_arg(1, Integer)
        def bar(self, integer=1):
            return 1 + integer

        @log_exceptions
        @assert_arg("integer", Integer)
        def bar2(self, integer=1):
            """
            Test docstring
            """
            return 1 + integer

        @log_exceptions
        @assert_prop("notExistingProp")
        @assert_arg("integer", Integer)
        def bar3(self, integer=1):
            """
            bar 3 documentation
            """
            return 1 + integer

    logging.basicConfig(level=logging.DEBUG)
    a = A()

    with pytest.raises(MyException) as e_info:
        a.raise_exc()
    with pytest.raises(ValidationError) as e_info:
        a.foo("reziu")
    with pytest.raises(ValidationError) as e_info:
        a.bar(integer="reziu")
    with pytest.raises(ValidationError) as e_info:
        a.bar2(integer="reziu")
    with pytest.raises(AttributeError) as e_info:
        a.bar3(integer=1)


if __name__ == "__main__":
    test_decorators()
# pytest.main(__file__)



# -*- coding: utf-8 -*-
"""
ngofile.exceptions
-----------------------
All exceptions used in the ngofile code base are defined here.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from jsonschema.exceptions import SchemaError
#from python_jsonschema_objects import ValidationError
from jsonschema.exceptions import FormatError, ValidationError


class ConversionError(ValueError):
    """
    Raised if an invalid value is detected in a validator
    """


class InvalidOperation(Exception):
    """
    Raised if an invalid operation is attempted
    """


class ContextError(ValueError):
    """
    Raised if an error happens during context evaluation
    """


class ExpressionError(ConversionError, ContextError):
    """
    Raised if an error happens during expression evaluation
    """


class InvalidValue(ConversionError, ValidationError):
    """
    Raised if an invalid value is detected in a validator
    """
    def __init__(self, *args, **kwargs):
        ValidationError.__init__(self, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
Unit tests for utilities

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import logging
logging.basicConfig(level=logging.INFO)

from ngomf.models import variable
from ngoschema.session import session_maker, scoped_session

Session = scoped_session(session_maker())


def test_session():
    s1 = Session()
    s2 = Session()
    v1 = variable.RealVariable(name="v1")
    v2 = variable.RealVariable(name="v2")
    s1.register(v1)
    s2.register(v2)


def test_global_session():
    pass


if __name__ == "__main__":
    test_session()



# -*- coding: utf-8 -*-
"""
Unit tests for utilities

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import logging
logging.basicConfig(level=logging.INFO)

from ngoschema.repositories import JsonFileRepository, YamlFileRepository
from ngoschema.models.documents import Document
#from ngomf.component import ComponentDefinition
from ngoci.models.project import Project

def test_json_handler():
    d1 = Document(filepath="/Users/cedric/Devel/python/python-ngomf/ngomf/models/Ngo/MoistAir/PhaseChange/MACND00.json")
    h1  = JsonFileRepository(instanceClass='ngosim.models.component.ComponentDefinition', document=d1)
    c1 = h1.load()

    d2 = Document(filepath="/Users/cedric/Devel/python/python-ngoci/tests/fixtures/projects.ngoprj")
    h2  = YamlFileRepository(instanceClass=Project, document=d2, many=True)
    c2s = h2.load()
    print(len(c2s))

if __name__ == "__main__":
    test_json_handler()



# *- coding: utf-8 -*-
"""
Utilities for objects queries

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import itertools
import operator
import re

from . import utils
from ngoschema.utils import get_descendant

_operators = [
    # operator library
    'lt',
    'le',
    'eq',
    'ne',
    'neq',
    'ge',
    'gt',
    'not',
    'contains',
    # redefined
    'in',
    'size',
    'intersects',
    # string and regex
    'ieq',
    'icontains',
    'startswith',
    'istartswith',
    'endswith',
    'iendswith',
    'regex'
]


def _comparable(obj):
    return obj
    from .protocol_base import ProtocolBase
    if isinstance(obj, ProtocolBase):
        #obj = obj[CN_KEY]
        obj = str(obj.canonicalName)
    if isinstance(obj, LiteralValue):
        return obj.for_json()
    return obj


def _sort_criteria(criteria):
    elts = criteria.split('__')
    ops = []
    for e in reversed(elts):
        if e not in _operators and not e.startswith('@'):
            break
        ops.insert(0, e)
        elts.pop()
    return elts, ops


def _apply_ops_test(ops, negate, a, b):
    def _apply_op(op, a, b):
        if op in ('startswith', 'endswith', 'ieq', 'istartswith', 'iendswith', 'icontains', 'regex'):
            if not utils.is_string(a):
                raise TypeError("%s requires a string (%s not a string)" % (op, a))
            if not utils.is_string(b):
                raise TypeError("%s requires a string (%s not a string)" % (op, b))
        if hasattr(operator, op):
            return getattr(operator, op)(a, b)
        if op == 'neq':
            return a != b
        if op == 'size':
            return len(a)
        if op.startswith('@'):
            return a[int(op[1:])]
        if op == 'in':
            return a in b
        if op == 'contains':
            return b in a
        if op == 'intersects':
            if not utils.is_sequence(a):
                raise TypeError("%s requires a sequence (%s not a sequence)" % (op, a))
            if not utils.is_sequence(b):
                raise TypeError("%s requires a sequence (%s not a sequence)" % (op, b))
            return set(a).intersection(set(b))
        if op == 'startswith':
            return a.startswith(b)
        if op == 'endswith':
            return a.endswith(b)
        if op == 'ieq':
            return re.match('^%s$' % b, a, re.IGNORECASE)
        if op == 'istartswith':
            return re.match('^%s' % b, a, re.IGNORECASE)
        if op == 'iendswith':
            return re.match('%s$' % b, a, re.IGNORECASE)
        if op == 'icontains':
            return re.search('%s' % b, a, re.IGNORECASE)
        if op == 'regex':
            return re.match(b, a)

    for op in ops:
        a = _apply_op(op, a, b)
    return not a if negate else a


class Filter(object):

    def __init__(self,
                 *attrs,
                 negate=False,
                 any_of=False,
                 **attrs_value):
        self.negate = negate
        self.anyOf = any_of
        self.attrs_value = attrs_value
        self.attrs = attrs

        self.attrs_ops = {}
        for k, v2 in attrs_value.items():
            ks, ops = _sort_criteria(k)
            ops_negate = 'not' in ops
            if ops_negate:
                ops.remove('not')
            self.attrs_ops[k] = (ks, ops, ops_negate)

    def __call__(self, obj):
        test = not self.anyOf
        for k, v2 in self.attrs_value.items():
            ks, ops, ops_negate = self.attrs_ops[k]
            o = get_descendant(obj, ks)
            if o is None:
                # breaking the look we never go in the for/ELSE statement where
                # an object is potentially yielded
                test = False
                break
            elif ops and utils.is_mapping(o):
                # check if it s not a child
                for op in ops:
                    o2 = get_descendant(o, op)
                    if o2:
                        o = o2
                        ks.append(ops.pop(0))
                    else:
                        test = False
                        break
            ops = ops or ['eq']
            v = _comparable(o)
            test2 = bool(_apply_ops_test(ops, ops_negate, v, v2))
            test = (test or test2) if self.anyOf else (test and test2)
            if self.anyOf:
                if test2:
                    break
            elif not test:
                break

        for k in self.attrs:
            ks = k.split('__')
            test2 = get_descendant(obj, ks) is not None
            test = (test or test2) if self.anyOf else (test and test2)
            if self.anyOf:
                if test2:
                    break
            elif not test:
                break
        # todo change
        if test is not self.negate:
            return True
        return False


def search_object(obj, path, *attrs, **attrs_value):
    """wrapper around dpath.util with filters on attributes presence and value"""
    import dpath.util, dpath.path
    afilter = Filter(*attrs, **attrs_value) if attrs or attrs_value else None
    separator = '/'
    globlist = dpath.util.__safe_path__(path, separator)
    for path in dpath.util._inner_search(obj, globlist, '/', dirs=True):
        val = get_descendant(obj, [p[0] for p in path])
        if afilter and afilter(val):
            yield (separator.join(map(str, dpath.path.paths_only(path))), val)


def search_object_non_rec(obj, path, *attrs, **attrs_value):
    res = next(search_object(obj, path, *attrs, **attrs_value))
    if res:
        p, e = res
        yield p, e
        # only next siblings and remaining next cousins, etc...
        p_cur, cur = p, e
        while cur and '/' in p_cur:
            if cur is obj:
                yield
            p_par = p_cur.rsplit('/', 1)[0]
            par = get_descendant(obj, p_par.split('/'))
            if utils.is_sequence(par):
                next_siblings = list(range(par.index(cur)+1, len(par)))
            else:
                next_siblings = {k for i, k in enumerate(par.keys()) if i > list(par.keys()).index(cur)}
            for s in next_siblings:
                for ps, pe in par[s].search_non_rec(path, *attrs, **attrs_value):
                    yield ps, pe
            p_cur, cur = p_par, cur._parent


class Query(object):

    def __init__(self, iterable, distinct=False, order_by=False, reverse=False):
        """
        :param iterable: iterator on which to perform the query
        :param distinct: flag to only return distinct objects
        :type distinct: boolean
        :param order_by: column to use for order
        :type order_by: string
        :param reverse: boolean to reverse order
        :type reverse: boolean
        """
        self._iterable = iterable
        self._distinct = distinct
        self._order_by = order_by
        self._reverse = reverse

    def _chain(self):
        return Query(self._iterable, self._distinct, self._order_by, self._reverse)

    @staticmethod
    def _filter_or_exclude(iterable,
                           *attrs,
                           negate=False,
                           any_of=False,
                           distinct=False,
                           **attrs_value):
        """
        Make a filter/exclude generator for an iterable. The flag `negate` allow to return
        the excluded objects corresponding to the criteria. This generator filters/excludes
        objects respecting ANY of the criteria (OR).

        Criteria on descendant attributes are done with the syntax `child__childOfChild`.
        The last elements of a criteria can be an operator as [lt, le, eq, ne, ge, gt, in] (
        default is `eq`)
        ex: `child__numElements__ge = 5` will select objects with an attribute named `child`
        with an attribute `numElements` greater or equal to 5

        :param iterable: iterable to process
        :param attrs: select/reject objects with given attributes defined
        :param attrs_value: select/objects objects with given attribute/value pairs
        """
        test_op = Filter(*attrs,
                           negate=negate,
                           any_of=any_of,
                           **attrs_value)
        seen = set()
        for obj in iterable:
            if not obj: # todo change to is None??
                continue
            if test_op(obj):
                if distinct:
                    comparable = _comparable(obj)
                    if comparable not in seen:
                        seen.add(comparable)
                    else:
                        continue
                yield obj

    @staticmethod
    def _filter_or_exclude__(iterable,
                               *attrs,
                               negate=False,
                               any_of=False,
                               distinct=False,
                               **attrs_value):

        seen = set()
        attrs_ops = {}
        for k, v2 in attrs_value.items():
            ks, ops = _sort_criteria(k)
            ops_negate = 'not' in ops
            if ops_negate:
                ops.remove('not')
            attrs_ops[k] = (ks, ops, ops_negate)

        for obj in iterable:
            if not obj:
                continue
            test = not any_of
            for k, v2 in attrs_value.items():
                ks, ops, ops_negate = attrs_ops[k]
                o = get_descendant(obj, ks)
                if o is None:
                    # breaking the look we never go in the for/ELSE statement where
                    # an object is potentially yielded
                    test = False
                    break
                elif ops and utils.is_mapping(o):
                    # check if it s not a child
                    for op in ops:
                        o2 = get_descendant(o, op)
                        if o2:
                            o = o2
                            ks.append(ops.pop(0))
                        else:
                            test = False
                            break
                ops = ops or ['eq']
                v = _comparable(o)
                test2 = bool(_apply_ops_test(ops, ops_negate, v, v2))
                test = (test or test2) if any_of else (test and test2)
                if any_of:
                    if test2:
                        break
                elif not test:
                    break

            for k in attrs:
                ks = k.split('__')
                test2 = get_descendant(obj, ks) is not None
                test = (test or test2) if any_of else (test and test2)
                if any_of:
                    if test2:
                        break
                elif not test:
                    break
            if test is not negate:
                if distinct:
                    comparable = _comparable(obj)
                    if comparable not in seen:
                        seen.add(comparable)
                    else:
                        continue
                yield obj

    def filter(self,
               *attrs,
               order_by=None,
               reverse=None,
               distinct=None,
               **attrs_value):
        return Query(
                (x for x in Query._filter_or_exclude(self._iterable, *attrs,
                                                    distinct=distinct or self._distinct,
                                                    **attrs_value)),
                order_by=order_by or self._order_by,
                reverse=reverse or self._reverse,
                distinct=distinct or self._distinct)

    def exclude(self,
               *attrs,
                order_by=None,
                reverse=None,
                distinct=None,
                **attrs_value):
        return Query(
                (x for x in Query._filter_or_exclude(self._iterable, *attrs,
                                                    negate=True,
                                                    distinct=distinct or self._distinct,
                                                    **attrs_value)),
                order_by=order_by or self._order_by,
                reverse=reverse or self._reverse,
                distinct=distinct or self._distinct)

    def filter_any_of(self,
               *attrs,
               order_by=None,
               reverse=None,
               distinct=None,
               **attrs_value):
        return Query(
                (x for x in Query._filter_or_exclude(self._iterable, *attrs,
                                                    any_of=True,
                                                    distinct=distinct or self._distinct,
                                                    **attrs_value)),
                order_by=order_by or self._order_by,
                reverse=reverse or self._reverse,
                distinct=distinct or self._distinct)

    def exclude_any_of(self,
                *attrs,
                order_by=None,
                reverse=None,
                distinct=None,
                **attrs_value):
        return Query(
                (x for x in Query._filter_or_exclude(self._iterable, *attrs,
                                                    any_of=True,
                                                    negate=True,
                                                    distinct=distinct or self._distinct,
                                                    **attrs_value)),
                order_by=order_by or self._order_by,
                reverse=reverse or self._reverse,
                distinct=distinct or self._distinct)

    def __iter__(self):
        return (x for x in self.result_cache)

    _result_cache = None
    @property
    def result_cache(self):
        if self._result_cache is None:
            self._result_cache = list(self._iterable)

            ob = self._order_by
            if ob:
                from .utils.utils import split_path
                ks = split_path(ob) if isinstance(ob, str) else ob
                self._result_cache = sorted(self._result_cache,
                                            lambda x: get_descendant(x, ks))
            if self._reverse:
                self._result_cache = reversed(self._result_cache)
        return self._result_cache

    def __getitem__(self, k):
        """Retrieve an item or slice from the set of results."""
        if not isinstance(k, (int, slice)):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        return self.result_cache[k]

    def __len__(self):
        return len(self.result_cache)

    def __contains__(self, obj):
        v = _comparable(obj)
        for o in self:
            if _comparable(o) == v:
                return True
        return False

    def union(self, *others):
        res = self.result_cache
        for o in others:
            res += o.result_cache
        return Query(res, self._distinct, self._order_by)

    def intersection(self, *others):
        res = self.result_cache
        for o in others:
            ores = o.result_cache
            res = [r for r in res if r in ores]
        return Query(res, self._distinct, self._order_by)

    def difference(self, *others):
        res = self.result_cache
        for o in others:
            ores = o.result_cache
            res = [r for r in res if r not in ores]
        return Query(res, self._distinct, self._order_by)

    def get(self, *attrs, **attrs_value):
        ret = list(Query._filter_or_exclude(self._iterable,
                                        *attrs,
                                        distinct=self._distinct,
                                        **attrs_value))
        if len(ret) == 0:
            raise Exception('Entry %s does not exist' % attrs_value)
        elif len(ret) > 1:
            import logging
            logging.error(ret)
            raise Exception('Multiple objects returned')
        return ret[0]

    def next(self, *attrs, **attrs_value):
        return next(
                Query._filter_or_exclude(self._iterable,
                                         *attrs,
                                         distinct=self._distinct,
                                         **attrs_value))

    def first(self):
        try:
            return self[0]
        except IndexError:
            return None

    def last(self):
        try:
            return self[-1]
        except IndexError:
            return None

    def reverse(self):
        ret = self._chain()
        ret._iterable = reversed(self.result_cache)
        return ret

    def count(self):
        return len(self)

    def isEmpty(self):
        return len(self) == 0

    def order_by(self, order_by, reverse=False):
        return Query(self._iterable, self._distinct, order_by, reverse=reverse)

    def all(self):
        """
        Return all results of query set
        """
        return self.result_cache



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" start-up script to launch all units tests """

import logging

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    import pytest

    pytest.main()



# -*- coding: utf-8 -*-
"""
Unit tests for utilities

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import logging
logging.basicConfig(level=logging.INFO)

import collections
import pathlib
from builtins import str

import pytest
import six
from future.utils import text_to_native_str

#from ngoschema.utils import str_utils
#from ngoschema import utils
#from ngoschema.exceptions import InvalidValue


def test_resolve_import():
    function = utils.import_from_string("os.path.dirname")
    module = utils.import_from_string("past.builtins")
    klass = utils.import_from_string("decimal.Decimal")
    assert utils.fullname(klass) == "decimal.Decimal"
    with pytest.raises(InvalidValue) as e_info:
        utils.import_from_string("not.an.importable")

    method = utils.import_from_string(
        "jsonschema.validators.RefResolver.resolve")
    assert utils.fullname(
        method) == "jsonschema.validators.RefResolver.resolve"
    # tested on temporary file containing a nested class, import and fullname were working
    # nested_class = utils.import_from_string('ngoschema.example_nested_class.Human.Head')
    # assert utils.fullname(nested_class) == 'ngoschema.example_nested_class.Human.Head'


def test_utils():
    assert utils.is_sequence([1, 2])
    assert utils.is_sequence((1, 2))
    assert utils.is_sequence(collections.deque([1, 2]))
    assert not utils.is_sequence(set([1, 2]))

    assert utils.is_mapping({"a": 1})
    assert utils.is_mapping(collections.OrderedDict(a=1))
    assert not utils.is_mapping(1)
    assert not utils.is_mapping("a")

    assert utils.is_collection([1, 2])
    assert utils.is_collection(set([1, 2]))
    assert utils.is_collection((1, 2))
    assert utils.is_collection(collections.deque([1, 2]))
    assert utils.is_collection({"a": 1})
    assert utils.is_collection(collections.OrderedDict(a=1))
    assert not utils.is_collection(1)
    assert not utils.is_collection("a")

    newstring = str("foo")
    basestring_ = text_to_native_str("foo")
    unicodestr = str("foo")
    path = pathlib.Path("a/dummy/path")

    assert utils.is_basestring(basestring_) == six.PY2
    assert not utils.is_basestring(unicodestr)
    assert not utils.is_basestring(newstring)
    assert not utils.is_basestring(path)

    assert utils.is_string(basestring_)
    assert utils.is_string(unicodestr)
    assert utils.is_string(newstring)
    assert not utils.is_string(path)

    assert utils.is_pattern("{{ typical pattern }}")
    assert not utils.is_pattern("not a pattern")

    assert utils.is_expr("`a+b")
    assert not utils.is_expr("{{ typical pattern }}")


def test_string_utils():
    assert (str_utils.multiple_replace("boa kills, is it a boa", {
        "boa": "python",
        "kills": "rocks"
    }) == "python rocks, is it a python")

    assert str_utils.split_string("one,two|three",
                                  ",|") == ["one", "two", "three"]

def test_jinja_tokens():
    from ngoschema.utils.jinja2 import get_jinja2_variables
    vars = get_jinja2_variables("{{ this.type }}-{{ this.name|lower|replace(' ','-') }}")
    print(vars)

def test_class_casted_as():

    from ngoschema.utils import class_casted_as

    class A:
        _a = 'world'
        @classmethod
        def f(cls):
            return 'hello %s!' % cls._a

    class B(A):
        _a = 'sunshine'
        @classmethod
        def f(cls):
            return 'good morning %s!' % cls._a

    assert B.f() == 'good morning sunshine!'
    assert class_casted_as(B, A).f() == 'hello sunshine!'


    class A:
        _a = 'world'
        @classmethod
        def f(cls):
            return A._f(cls)

        @staticmethod
        def _f(cls):
            return 'hello %s!' % cls._a

    class B(A):
        _a = 'sunshine'
        @classmethod
        def f(cls):
            return B._f(cls)

        @staticmethod
        def _f(cls):
            return 'good morning %s!' % cls._a

    assert B.f() == 'good morning sunshine!'
    assert A._f(B) == 'hello sunshine!'



if __name__ == "__main__":
    test_class_casted_as()
    test_jinja_tokens()
    #break
    #test_resolve_import()
    #test_utils()
    #test_string_utils()



import pytest

from future.utils import with_metaclass
from ngoschema.exceptions import InvalidValue
from ngoschema.loaders.schemas import load_schema
from ngoschema.protocols import SchemaMetaclass, split_cname

split_cname('asd[1][2]')

load_schema({
    '$id': 'A',
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'a': {'type': 'string'},
    }
})

load_schema({
    '$id': 'B',
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'b': {'type': 'integer'},
    }
})

load_schema({
    '$id': 'BextA',
    'type': 'object',
    'extends': ['A'],
    'additionalProperties': False,
    'properties': {
        'b': {'type': 'integer'},
    }
})


def test_object_protocol():
    class A(with_metaclass(SchemaMetaclass)):
        _id = 'A'

    class B(with_metaclass(SchemaMetaclass)):
        _id = 'B'

    class AB(with_metaclass(SchemaMetaclass, A, B)):
        a = '1'

    a = A(a=1)
    assert a.a == '1'
    with pytest.raises(AttributeError) as e_info:
        a.b = 1
    a.a = 2
    assert a.a == '2', a.a

    b = B(b=1)
    assert b.b == 1, b.b
    with pytest.raises(AttributeError) as e_info:
        b.a = 1

    ab = AB(b=1)
    assert ab.a == '1', ab.a
    assert ab.b == 1, ab.b
    ab.c = 1  # additional properties default is True, not redefined in AB
    assert ab.c == 1

    class BA(with_metaclass(SchemaMetaclass)):
        _id = 'BextA'

    ba = BA(a=1)
    assert ba.a == '1', ba.a


def test_call_order():
    # test smart lazy loading and property evaluation order
    class Foo(with_metaclass(SchemaMetaclass)):
        _lazyLoading = True
        _schema = {
            'type': 'object',
            'additionalProperties': False,
            'required': ['b'],
            'readOnly': ['c'],
            'properties': {
                'a': {'type': 'string', 'default': '{{b}}'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer', 'default': '`int(a)+1', 'maximum': 2}
            }
        }

    foo = Foo(b=1)
    # access attributes through descriptors ensuring conversion and validation
    assert foo.c == 2  # properly evaluate expression with with a not yet evaluated
    assert foo.a == '1'  # properly casted to string
    with pytest.raises(AttributeError) as e_info:
        foo.c = 2  # c is read only
    assert len(foo) == 3  # behaves like normal dict
    assert foo.a == foo['a']
    for k, v in foo.items():
        assert getattr(foo, k) == v
    # set a valid value, which makes another one invalid
    foo.b = 2  # ok so far
    with pytest.raises(InvalidValue) as e_info:
        foo.c  # invalid value as maximum is 2 and evaluation is 3
    foo.b = '1'  # reset to a compatible value for c, setting an integer convert from string
    assert foo['b'] == foo.b == 1
    foo.do_validate()
    # create a json dict with filtering options
    foo2 = foo.do_serialize(no_defaults=False, excludes=Foo._readOnly)
    assert foo2['a'] == foo.a
    assert 'c' not in foo2
    assert 'c' in foo.do_serialize(no_defaults=False)


def test_schema_mro():
    import ngosim
    from ngoschema.types import type_builder
    mro = type_builder.schema_mro('https://numengo.org/ngosim#$defs/variable/$defs/EnumerationVariable')
    assert mro


def test_repr():
    class Obj(with_metaclass(SchemaMetaclass)):
        _schema = {
            'type': 'object',
            'properties': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
            }
        }

    print(repr(Obj))
    fa = Obj(a=1)
    fb = Obj(b=1)
    print(repr(fa))
    print(repr(fb))
    print(str(Obj))
    print(str(fa))
    print(str(fb))

    class Arr(with_metaclass(SchemaMetaclass)):
        _schema = {
            'type': 'array',
            'items': {
                'type': 'string'
            }
        }

    #print(repr(Arr))
    a1 = Arr(['a', 'b'])
    print(repr(a1))
    print(str(a1))


def test_repr2():
    from ngoschema.models.files import Document
    print(repr(Document._properties))



if __name__ == "__main__":
    #test_repr2()
    test_object_protocol()
    test_call_order()
    test_repr()

    #test_schema_mro()



# -*- coding: utf-8 -*-
""" Unit tests for inspection utilities

test_inspect.py - created on 22/05/2018
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3  """

from __future__ import print_function
from __future__ import unicode_literals

from ngoschema import load_module_schemas


def test_FunctionInspector():
    from ngoinsp.inspectors.inspect_objects import FunctionInspector
    fi = FunctionInspector(load_module_schemas)
    assert fi.parameters[0].name == "module"
    assert fi.parameters[1].name == "schemas_store"
    assert fi.parameters[0].doc
    assert fi.returns
    assert len(fi.parameters) == 2
    assert not fi.keywords
    assert not fi.varargs
    assert not fi.decorators


def test_ClassInspector():
    from ngoinsp.inspectors.inspect_objects import ClassInspector
    from ngoschema import ProtocolBase

    ci = ClassInspector(ProtocolBase)
    assert ci.methods["set_configfiles_defaults"].shortDescription
    assert ci.methods["set_configfiles_defaults"].parameters[0].name == "overwrite"
    assert not ci.methods["set_configfiles_defaults"].parameters[0].default
    assert ci.methods["set_configfiles_defaults"].parameters[0].doc


if __name__ == "__main__":
    test_FunctionInspector()
    test_ClassInspector()



import os.path
import pathlib
from datetime import date
from datetime import datetime
from datetime import time

import arrow
import pytest  # noqa
import python_jsonschema_objects as pjo

from ngoschema import get_resolver
from ngoschema.classbuilder import ClassBuilder


def test_path():
    schema = {
        "title": "Path Example",
        "type": "object",
        "properties": {
            "listPath": {
                "type": "array",
                "items": {
                    "type": "path"
                }
            },
            "anyPath": {
                "type": "path"
            },
            "existingPath": {
                "type": "path",
                "isPathExisting": True
            },
            "dirPath": {
                "type": "path",
                "isPathDir": True
            },
        },
    }
    builder = ClassBuilder(get_resolver())
    PE = builder.construct("PathExample", schema)

    fp = os.path.abspath(__file__)
    dp = os.path.dirname(fp)

    pe = PE()
    pe.anyPath = "really any path"
    assert isinstance(pe.anyPath._value, pathlib.Path)
    pe.anyPath.exists()
    with pytest.raises(pjo.ValidationError):
        pe.existingPath = "really any path"
    pe.existingPath = fp
    assert isinstance(pe.existingPath._value, pathlib.Path)
    with pytest.raises(pjo.ValidationError):
        pe.dirPath = fp
    pe.dirPath = dp
    assert isinstance(pe.existingPath._value, pathlib.Path)
    pe.listPath = ["really any path", fp, dp]
    for p in pe.listPath:
        assert isinstance(p._value, pathlib.Path)


def test_datetime():
    schema = {
        "title": "Date Time Example",
        "type": "object",
        "properties": {
            "date": {
                "type": "date"
            },
            "time": {
                "type": "time"
            },
            "datetime": {
                "type": "datetime"
            },
        },
    }

    builder = ClassBuilder(get_resolver())
    DTE = builder.construct("DateTimeExample", schema)

    d = date(2018, 5, 26)
    a_d = arrow.get(d)
    dt_d = a_d.datetime

    t = time(11, 11, 11)
    dt_t = datetime(1, 1, 1, 11, 11, 11)
    a_t = arrow.get(dt_t)

    dt = datetime(2018, 5, 26, 11, 11, 11)
    a_dt = arrow.get(dt)

    dte = DTE()

    # takes time
    dte.time = t
    # takes string formatted as time
    dte.time = "11:11"
    assert dte.time.hour == 11
    assert dte.time.minute == 11
    assert isinstance(dte.time._value, time)
    # takes datetime with no date
    dte.time = dt_t
    assert isinstance(dte.time._value, time)
    # takes arrow with no date
    dte.time = a_t
    assert isinstance(dte.time._value, time)
    # doesn t take date
    with pytest.raises(pjo.ValidationError):
        dte.time = d
    # doesn t take datetime with date
    with pytest.raises(pjo.ValidationError):
        dte.time = dt

    # takes date
    dte.date = d
    # takes string formatted as date
    dte.date = "2018/05/26"
    assert dte.date.year == 2018
    assert dte.date.month == 5
    assert dte.date.day == 26
    assert isinstance(dte.date._value, date)
    dte.date = "26/05/2018"
    assert dte.date.year == 2018
    assert dte.date.month == 5
    assert dte.date.day == 26
    # takes datetime with no time
    dte.date = dt_d
    assert not isinstance(dte.date, date)
    assert isinstance(dte.date._value, date)
    # takes arrow with no time
    dte.date = a_d
    assert isinstance(dte.date._value, date)
    # doesn t take time
    with pytest.raises(pjo.ValidationError):
        dte.date = t
    # take only date from datetime with time
    dte.date = dt
    assert isinstance(dte.date._value, date)

    # takes date
    dte.datetime = d
    assert isinstance(dte.datetime._value, arrow.Arrow)
    # takes datetime
    dte.datetime = dt
    assert isinstance(dte.datetime._value, arrow.Arrow)
    # takes arrow
    dte.datetime = a_dt
    assert isinstance(dte.datetime._value, arrow.Arrow)

def test_literal():
    from ngoschema import Path
    p = Path('/Users/cedric/Devel/admin/NUMENGO/NGOSCHEMA/url.mm')
    assert p.to_uri()

if __name__ == "__main__":
    #test_path()
    #test_datetime()
    test_literal()



import pytest
from ngoschema.exceptions import *
from ngoschema import types
import ngoschema.types.symbols as symbols

from collections import OrderedDict
import pathlib
import decimal
from urllib.parse import ParseResult
from datetime import datetime, date, time
import arrow


def test_base():
    i = types.Integer(maximum=0)
    assert repr(i) == "ngoschema.types.numerics.Integer(type='integer', maximum=0)"
    assert str(i) == "<ngoschema.types.numerics.Integer type='integer' maximum=0>"
    assert types.Integer()(1) == 1
    assert types.Integer()("1") == 1
    assert types.Integer()("`1+1") == 2
    assert types.Integer()("{{a}}", a=1) == 1
    assert types.Integer(minimum=0, maximum=2)(1) == 1
    with pytest.raises(InvalidValue) as e_info:
        types.Integer(maximum=0)(1) == 1

    assert types.String()(1) == "1"
    assert types.String()("hello world!") == "hello world!"
    assert types.String()("hello {{name}}!", name='world') == 'hello world!'
    assert types.String()("hello {{name}}!", name='world', raw_literals=True) == "hello {{name}}!"
    with pytest.raises(ValidationError) as e_info:
        assert types.String(maxLength=5)("hello {{name}}!", name='world') == 'hello world!'

    assert types.Boolean.check('TRUE', convert=True)
    assert types.Boolean.check('false', convert=True)
    assert not types.Boolean.check('not a boolean', convert=True)
    assert types.Boolean()(True) is True
    assert types.Boolean()(False) is False
    assert types.Boolean()('true', convert=True) is True
    assert types.Boolean()('TRUE', convert=True) is True
    assert types.Boolean()('false', convert=True) is False

    assert types.Number()(2.3) == pytest.approx(decimal.Decimal('2.3')), types.Number()(2.3)
    with pytest.raises(InvalidValue) as e_info:
        assert types.Number(maximum=2)(2.3)

    t = types.Type(type='string')
    assert repr(t) == "ngoschema.types.type.Type(type='string')"
    assert str(t) == "<ngoschema.types.type.Type type='string'>"
    assert types.Type(type='string')(1) == '1', types.Type(type='string')(1)
    assert types.Type(type='integer', minimum=0)(1) == 1
    assert types.Type(type='number')(1) == 1
    assert isinstance(types.Type(type='boolean')(1), bool)


def test_complex():
    Path = types.Path()
    PathFileExists = types.PathFileExists()
    assert Path(__file__) == pathlib.Path(__file__)
    assert PathFileExists(__file__) == pathlib.Path(__file__)
    assert Path('{{user_directory}}', user_directory='.') == pathlib.Path('.')
    assert Path(pathlib.Path('{{user_directory}}'), user_directory='.') == pathlib.Path('.')

    Datetime = types.Datetime()
    dt = datetime(2018, 5, 26, 11, 11, 11)
    a_dt = arrow.get(dt)

    assert Datetime(dt), dt
    assert isinstance(Datetime(dt), arrow.Arrow)
    assert Datetime(a_dt), a_dt
    assert isinstance(Datetime(a_dt), arrow.Arrow)

    Time = types.Time()
    t = time(11, 11, 11)
    dt_t = datetime(1, 1, 1, 11, 11, 11)
    a_t = arrow.get(dt_t)
    assert Time(t) == t, t
    assert Time(dt_t) == t, dt_t
    assert Time(a_t) == t, a_t

    Date = types.Date()
    d = date(2018, 5, 26)
    a_d = arrow.get(d)
    dt_d = a_d.datetime
    assert Date(d) == d, d
    assert Date(a_d) == d, a_d
    assert Date(dt_d) == d, dt_d


def test_array():
    ArrayString = types.Array(items={'type': 'string'})
    ArrayInteger = types.Array(items={'type': 'integer'})
    ArrayInteger5 = types.Array(items={'type': 'integer'}, minItems=5, maxItems=5)
    #print("|%s|" % repr(ArrayString))
    #print("|%s|" % str(ArrayString))
    #print("|%s|" % repr(ArrayInteger5))
    assert repr(ArrayString) == "ngoschema.types.array.Array(type='array', items={'type': 'string'})"
    assert str(ArrayString) == "<ngoschema.types.array.Array type='array' items{1}>"
    assert repr(ArrayInteger5) == "ngoschema.types.array.Array(type='array', items={'type': 'integer'}, minItems=5, maxItems=5)", repr(ArrayInteger5)
    assert ArrayInteger([1, '2', '{{a}}', '`a'], a=1) == [1, 2, 1, 1], ArrayInteger([1, '2', '{{a}}', '`a'], a=1)
    assert ArrayString([1, '2', '{{a}}', '`a'], a=1) == ['1', '2', '1', '1']
    #with pytest.raises(InvalidValue) as e_info:
    #    ArrayInteger5([1, '2', '{{a}}', '`a'], a=1)
    assert ArrayInteger5.default() == [None, None, None, None, None], ArrayInteger5.default()
    # below should be the result if the integer had a default value set to 0
    #assert ArrayInteger5._default() == [0, 0, 0, 0, 0], ArrayInteger5._default()


def test_object():
    from ngoschema.types import Object
    o = {'a': '1', 'b': 1}
    obj = Object()
    #assert obj(o) == o, obj(o)
    #assert obj(**o) == o, obj(o)

    obj2 = Object(properties={'a': types.Integer})
    assert obj2(o)['a'] == 1
    obj3 = Object(required=['c'])
    with pytest.raises(InvalidValue) as e_info:
        obj3(o, convert=False)
    assert obj3(o, convert=True) == {'a': '1', 'b': 1, 'c': None}
    assert obj3.default() == OrderedDict(c=None)


def test_canonical_name():
    from ngoschema.contexts import ObjectProtocolContext as Context
    from ngoschema.types.foreign_key import CanonicalName
    b = {'name': 'b'}
    a = 'a value'
    context = Context({
        'A': {
            'a': a,
            0: b
        }
    })
    cn = CanonicalName()
    assert cn.resolve('A.b', context) == b
    assert cn.resolve('A.a', context) == 'a value'
    with pytest.raises(InvalidValue) as e_info:
        cn.resolve('A.c')


if __name__ == '__main__':
    #test_canonical_name()
    test_object()
    test_base()
    test_array()
    test_complex()



from ngoschema.models import EntityWithMetadata


def test_fkey_relationship():

    A = EntityWithMetadata(name="A")
    B = EntityWithMetadata(name="B")
    C = EntityWithMetadata(name="C")

    assert str(A.canonicalName) == 'A', str(A.canonicalName)
    assert str(B.canonicalName) == 'B', str(B.canonicalName)

    # set a child through parent member
    B._parent = A
    assert B._parent is A, B._parent
    # check canonical names
    assert str(A.canonicalName) == 'A', str(A.canonicalName)
    assert str(B.canonicalName) == 'A.B', str(B.canonicalName)
    # check children
    assert len(A.children)==1, len(A.children)
    assert A.children[0].ref is B, A.children[0].ref

    # set another child through parent member
    C._parent = A
    assert len(A.children)==2
    assert len(B.children)==0
    assert len(C.children)==0

    # remove reference
    C._parent = None
    assert len(A.children)==1

    # test adding a children
    A.children.append(C)
    assert len(A.children)==2
    assert C._parent.ref is A

    # change name and check propagation
    A.name = 'AA'
    assert str(A.name) == 'AA'
    assert A.iname == 'AA'
    assert str(B._parent) == 'AA', str(B._parent)
    assert str(A.canonicalName) == 'AA'
    assert str(B.canonicalName) == 'AA.B'

    # check override of canonical name and its propagation
    A.canonicalName = 'AAA.AA'
    assert B.canonicalName == 'AAA.AA.B'

    C._parent = None
    assert len(A.children)==1, len(A.children)

    B._parent = None
    assert B._parent.ref is None
    assert B.canonicalName == 'B'

if __name__ == "__main__":
    test_fkey_relationship()



# -*- coding: utf-8 -*-

__author__ = """Cedric ROMAN"""
__email__ = "roman@numengo.com"
__version__ = "__version__ = '0.4.6'"

# load settings
from simple_settings import LazySettings
settings = LazySettings('ngoschema.config.settings', 'NGOSCHEMA_.environ')

# register module and load schemas
from .loaders import register_module
register_module('ngoschema')

# create a default context
#from .contexts import Context

#DEFAULT_CONTEXT = Context(**settings.DEFAULT_CONTEXT)
#APP_CONTEXT = DEFAULT_CONTEXT.create_child(settings)

from .exceptions import InvalidOperation, SchemaError, ValidationError
from .managers import *
from .protocols import *
from .repositories import *
from .query import Query, Filter
from .registries import serializers_registry, transformers_registry, repositories_registry


__all__ = [
    'settings',
    # exceptions
    'SchemaError',
    'ValidationError',
    'InvalidOperation',
    # loaders
    'register_module',
    # builder and protocol
    'TypeBuilder',
    'type_builder',
    'NamespaceManager',
    'with_metaclass',
    'SchemaMetaclass',
    # literals
    # query
    'Query',
    'Filter',
    # repository
    'repositories_registry',
]



# -*- coding: utf-8 -*-
"""

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import pytest

doc1 = {
    "filepath": r"D:\CODES\projects.ngoprj",
    "author": "doc1",
    "description": "a description",
    "keywords": ["cpp"]
}

doc2 = {
    "filepath": r"D:\CODES\ngoci.py.ngocc",
    "author": "doc2",
    "description": "a description",
    "keywords": ["python", "ci"]
}

doc3 = {
    "filepath": r"D:\CODES\ngoschema.py.ngocc",
    "author": "_doc3",
    "keywords": ["schemas", "json"]
}

from ngoschema.repositories import Repository
from ngoschema.models import Document


def test_factory_and_query():
    doc_loader = Repository(instanceClass='ngoschema.models.Document', many=True)
    doc_loader.register(Document(**doc1))
    doc_loader.register(Document(**doc2))
    doc_loader.register(Document(**doc3))


    assert len(doc_loader.filter(filepath__suffix__not=".ngoprj").all())==2
    l2 = list(doc_loader.filter(filepath__suffix__not=".ngoprj"))
    assert len(l2)==2, l2

    assert len(doc_loader.filter(keywords__size__ge=2, load_lazy=True)) == 2
    assert len(doc_loader.filter(any_of=True, filepath__suffix=".ngoprj", author__endswith="2", load_lazy=True)) == 2
    assert len(doc_loader.filter(keywords__intersects=["schemas", "python"])) == 2
    with pytest.raises(TypeError):
        assert len(doc_loader.filter(keywords__intersects="python")) == 2, len(doc_loader.filter(keywords__intersects="python"))
    #assert len(doc_loader.exclude(filepath__suffix=".ngoprj")) == 2
    assert len(doc_loader.filter(filepath__suffix=".ngoprj")) == 1
    assert len(doc_loader.filter(filepath__suffix__not=".ngoprj")) == 2


    assert len(doc_loader.filter(filepath__suffix__not=".ngoprj", author__istartswith="doc")) == 1
    assert len(doc_loader.filter(author__istartswith="doc")) == 2
    assert len(doc_loader.filter(author__icontains="doc")) == 3
    assert len(doc_loader.filter("description")) == 2
    #assert len(doc_loader.exclude("description")) == 1
    #assert len(doc_loader.filter_any_of("description", filepath__suffix__not=".ngoprj")) == 3
    #assert len(doc_loader.exclude_any_of(filepath__suffix=".ngoprj", author__endswith="2", load_lazy=True)) == 1
    #assert len(doc_loader.exclude_any_of("description", filepath__suffix=".ngoprj", author__endswith="3")) == 0


if __name__ == "__main__":
    test_factory_and_query()



# *- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

from .utils.utils import GenericClassRegistry


transformers_registry = GenericClassRegistry()

deserializers_registry = GenericClassRegistry()

serializers_registry = GenericClassRegistry()

repositories_registry = GenericClassRegistry()




# -*- coding: utf-8 -*-
""" Unit tests for inspection utilities

test_inspect.py - created on 22/05/2018
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3  """

from __future__ import print_function
from __future__ import unicode_literals


def test_doc_parsing():

    ds = """
        Create a new file with a given template

        This is a long description of the function.
    
        :param filetype: type of file
        :type filetype: type1, type2, type3
        :param name: name of the file
        :type name: string

        :rtype: bool
        """
    from ngoinsp.inspectors.doc_rest_parser import parse_docstring
    doc = parse_docstring(ds)
    assert doc["params"]["filetype"]["doc"] == "type of file"
    assert doc["params"]["filetype"]["type"] == "type1, type2, type3"
    assert len(doc["params"]) == 2
    assert doc["returns"] == "bool"


def test_type_parsing():
    from ngoinsp.inspectors.doc_rest_parser import parse_type_string as pts
    t1 = pts("str")
    assert t1["type"] == "string"

    t2 = pts("[str,int]")
    assert t2["type"] == ["string", "integer"]

    t3 = pts("array, items:{type:str,format:regex}")
    assert t3["type"] == "array"
    assert t3["items"]["type"] == "string"
    assert t3["items"]["format"] == "regex"

    # replace = with :
    t4 = pts("dict, items={type=str}")
    assert t4["type"] == "object"
    assert t4["items"]["type"] == "string"

    t5 = pts("enum:[A,B,C]")
    assert t5["enum"] == ["A", "B", "C"]


if __name__ == "__main__":
    test_doc_parsing()
    test_type_parsing()



# -*- coding: utf-8 -*-
"""
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import json
import time

from ngofile.list_files import list_files

#from ngoschema.canonical_name import _doc_cn_store
from ngoschema.models.documents import Document
from ngoschema.models.documents import get_document_registry
from ngoschema.resolvers.uri_resolver import _uri_doc_store

dirpath = "/Users/cedric/Devel/python/python-ngomf/ngomf/models/Ngo"

def test_DocumentRegistry2():
    s = time.time()
    ls = list(
        list_files(
            dirpath,
            includes='*.json',
            recursive=True))
    print('list_files', time.time() - s, len(ls))
    s = time.time()
    os = [
        Document(filepath=f, validate=True) for f in list_files(
            dirpath,
            recursive=True)
    ]
    print('list_files no lazy', time.time() - s, len(os))
    s = time.time()
    os = [
        Document(filepath=f, validate=False)
        for f in list_files(
            dirpath,
            recursive=True)
    ]
    print('list_files lazy no validation', time.time() - s, len(os))
    s = time.time()
    os = [
        Document(filepath=f, validate=False) for f in list_files(
            dirpath,
            recursive=True)
    ]
    print('list_files lazy', time.time() - s, len(os))

    s = time.time()
    doc_reg = get_document_registry()
    doc_reg.register_from_directory(
        dirpath,
        includes='*.json',
        recursive=True,
    )
    print(len(doc_reg.registry.values()))
    print('register', time.time() - s)
    doc = list(doc_reg.registry.values())[0]
    assert doc.loaded == False
    doc.deserialize()
    assert doc.loaded == True
    s = time.time()
    [doc.deserialize(deserializers=json) for doc in doc_reg]
    print('load', time.time() - s)
    print('uri store', len(_uri_doc_store))


def test_DocumentRegistry():
    s = time.time()
    doc_reg = get_document_registry()
    doc_reg.register_from_directory(
        dirpath,
        includes='*.json',
        recursive=True)
    print('register and load', time.time() - s)
    print('uri store', len(_uri_doc_store))
    query = doc_reg.query(filepath__name='NGOMED00.json')[:]
    assert len(query) == 1


if __name__ == "__main__":
    #test_DocumentRegistry2()
    test_DocumentRegistry()




# -*- coding: utf-8 -*-
"""

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import weakref
import functools

from sqlalchemy.util import ScopedRegistry, ThreadLocalRegistry

from . import utils
from .decorators import assert_arg
from .protocols import ObjectProtocol, ArrayProtocol, SchemaMetaclass, with_metaclass
from .types import Tuple, Array
from .query import Query

_sessions = weakref.WeakValueDictionary()

_new_sessionid = utils.threadsafe_counter()


def _state_session(state):
    """Given an :class:`.InstanceState`, return the :class:`.Session`
        associated, if any.
    """
    if state.session_id:
        try:
            return _sessions[state.session_id]
        except KeyError:
            pass
    return None


class Session(with_metaclass(SchemaMetaclass)):
    _id = "https://numengo.org/ngoschema#/$defs/session/$defs/Session"

    def __init__(self, **kwargs):
        self._resolve_cname = functools.lru_cache(1024)(self._resolve_cname_cached)
        ObjectProtocol.__init__(self, value=kwargs)
        self._new = {}  # InstanceState->object, strong refs object
        self._deleted = {}  # same
        self._hash_key = _new_sessionid()
        _sessions[self._hash_key] = self

    def bind_repo(self, repo):
        self.repositories.append(repo)
        repo._session = self

    def get_or_create_repo(self, name, repo_registry=None):
        from .repositories import repositories_registry
        repo_registry = repo_registry or repositories_registry
        for r in self.repositories:
            if r.__class__.__name__ == name:
                return r
        # not found, create binded repository
        r = repo_registry.get(name)
        if r:
            return r(session=self)
        raise ValueError('no repository %s in registry %.' % (name, repo_registry))

    def get_or_create_repo_by_class(self, instance_class, repo_registry=None):
        from .repositories import repositories_registry
        repo_registry = repo_registry or repositories_registry
        for r in self.repositories:
            if r.instanceClass is instance_class:
                return r
        # not found, create binded repository
        # look first in repo registry
        for kr, r in repo_registry.items():
            if r.instanceClass is instance_class:
                return r(session=self)  # repo is binded at initialization
        raise ValueError('no instance_class %s in registry %.' % (instance_class, repo_registry))

    def resolve_cname(self, cname):
        return self._resolve_cname(cname)

    def _resolve_cname_cached(self, cname):
        from .models.instances import Entity
        cns = cname.split('.')
        rn = cns[0]
        cn = cns[1:]
        for repo in [r for r in self._repos if issubclass(r.instanceClass, Entity)]:
            if rn in repo:
                v = repo.resolve_fkey(rn)
                return v if not cn else v.resolve_cname(cn)
        raise Exception("Impossible to resolve '%s'" % cname)

    @assert_arg(1, Tuple, strDelimiter=',')
    def resolve_fkey(self, keys, object_class):
        for repo in [r for r in self.repositories if issubclass(r.instanceClass, object_class)]:
            if keys in repo:
                return repo.resolve_fkey(keys)
        else:
            raise Exception("Impossible to resolve '%s'" % keys)

    def query(self, *attrs, order_by=False, **attrs_value):
        """
        Make a `Query` on registered documents
        """
        return Query(self._chained.values()).filter(
            *attrs, order_by=order_by, **attrs_value)

    def commit(self):
        pass

    def close(selfself):
        pass

    def _add_bind(self, key, bind):
        self.__bind[key] = bind

    def query(self, *entities, **kwargs):
        """Return a new :class:`.Query` object corresponding to this
        :class:`.Session`."""

        return self._query_cls(entities, self, **kwargs)

    def expire(self, instance):
        """Expire the attributes on an instance.

        Marks the attributes of an instance as out of date. When an expired
        attribute is next accessed, a query will be issued to the
        :class:`.Session` object's current transactional context in order to
        load all expired attributes for the given instance.   Note that
        a highly isolated transaction will return the same values as were
        previously read in that same transaction, regardless of changes
        in database state outside of that transaction"""
        pass

    def add(self, instance):
        pass

    def expunge(self, instance):
        pass

    def delete(self, instance):
        pass

    def merge(self, instance, load=True):
        pass


class session_maker(object):
    """A configurable :class:`.Session` factory.

    The :class:`.sessionmaker` factory generates new
    :class:`.Session` objects when called, creating them given
    the configurational arguments established here.

    e.g.::

        # global scope
        Session = sessionmaker(autoflush=False)

        # later, in a local scope, create and use a session:
        sess = Session()

    Any keyword arguments sent to the constructor itself will override the
    "configured" keywords::

        Session = sessionmaker()

        # bind an individual session to a connection
        sess = Session(bind=connection)

    The class also includes a method :meth:`.configure`, which can
    be used to specify additional keyword arguments to the factory, which
    will take effect for subsequent :class:`.Session` objects generated.
    This is usually used to associate one or more :class:`.Engine` objects
    with an existing :class:`.sessionmaker` factory before it is first
    used::

        # application starts
        Session = sessionmaker()

        # ... later
        engine = create_engine('sqlite:///foo.db')
        Session.configure(bind=engine)

        sess = Session()

    .. seealso:

        :ref:`session_getting` - introductory text on creating
        sessions using :class:`.sessionmaker`.

    """

    def __init__(
        self,
        bind=None,
        class_=Session,
        autoflush=True,
        autocommit=False,
        expire_on_commit=True,
        info=None,
        **kw
    ):
        r"""Construct a new :class:`.sessionmaker`.

        All arguments here except for ``class_`` correspond to arguments
        accepted by :class:`.Session` directly.  See the
        :meth:`.Session.__init__` docstring for more details on parameters.

        :param bind: a :class:`.Engine` or other :class:`.Connectable` with
         which newly created :class:`.Session` objects will be associated.
        :param class\_: class to use in order to create new :class:`.Session`
         objects.  Defaults to :class:`.Session`.
        :param autoflush: The autoflush setting to use with newly created
         :class:`.Session` objects.
        :param autocommit: The autocommit setting to use with newly created
         :class:`.Session` objects.
        :param expire_on_commit=True: the expire_on_commit setting to use
         with newly created :class:`.Session` objects.
        :param info: optional dictionary of information that will be available
         via :attr:`.Session.info`.  Note this dictionary is *updated*, not
         replaced, when the ``info`` parameter is specified to the specific
         :class:`.Session` construction operation.

         .. versionadded:: 0.9.0

        :param \**kw: all other keyword arguments are passed to the
         constructor of newly created :class:`.Session` objects.

        """
        kw["bind"] = bind
        kw["autoflush"] = autoflush
        kw["autocommit"] = autocommit
        kw["expire_on_commit"] = expire_on_commit
        if info is not None:
            kw["info"] = info
        self.kw = kw
        # make our own subclass of the given class, so that
        # events can be associated with it specifically.
        self.class_ = type(class_.__name__, (class_,), {})

    def __call__(self, **local_kw):
        """Produce a new :class:`.Session` object using the configuration
        established in this :class:`.sessionmaker`.

        In Python, the ``__call__`` method is invoked on an object when
        it is "called" in the same way as a function::

            Session = sessionmaker()
            session = Session()  # invokes sessionmaker.__call__()

        """
        for k, v in self.kw.items():
            if k == "info" and "info" in local_kw:
                d = v.copy()
                d.update(local_kw["info"])
                local_kw["info"] = d
            else:
                local_kw.setdefault(k, v)
        return self.class_(**local_kw)

    def configure(self, **new_kw):
        """(Re)configure the arguments for this sessionmaker.

        e.g.::

            Session = sessionmaker()

            Session.configure(bind=create_engine('sqlite://'))
        """
        self.kw.update(new_kw)


def close_all_sessions():
    for sess in _sessions.values():
        sess.close()


class scoped_session(object):
    """Provides scoped management of :class:`.Session` objects.

    See :ref:`unitofwork_contextual` for a tutorial.

    """

    session_factory = None
    """The `session_factory` provided to `__init__` is stored in this
    attribute and may be accessed at a later time.  This can be useful when
    a new non-scoped :class:`.Session` or :class:`.Connection` to the
    database is needed."""

    def __init__(self, session_factory, scopefunc=None):
        """Construct a new :class:`.scoped_session`.

        :param session_factory: a factory to create new :class:`.Session`
         instances. This is usually, but not necessarily, an instance
         of :class:`.sessionmaker`.
        :param scopefunc: optional function which defines
         the current scope.   If not passed, the :class:`.scoped_session`
         object assumes "thread-local" scope, and will use
         a Python ``threading.local()`` in order to maintain the current
         :class:`.Session`.  If passed, the function should return
         a hashable token; this token will be used as the key in a
         dictionary in order to store and retrieve the current
         :class:`.Session`.

        """
        self.session_factory = session_factory

        if scopefunc:
            self.registry = ScopedRegistry(session_factory, scopefunc)
        else:
            self.registry = ThreadLocalRegistry(session_factory)

    def __call__(self, **kw):
        r"""Return the current :class:`.Session`, creating it
        using the :attr:`.scoped_session.session_factory` if not present.

        :param \**kw: Keyword arguments will be passed to the
         :attr:`.scoped_session.session_factory` callable, if an existing
         :class:`.Session` is not present.  If the :class:`.Session` is present
         and keyword arguments have been passed,
         :exc:`~sqlalchemy.exc.InvalidRequestError` is raised.

        """
        if kw:
            if self.registry.has():
                raise sa_exc.InvalidRequestError(
                    "Scoped session is already present; "
                    "no new arguments may be specified."
                )
            else:
                sess = self.session_factory(**kw)
                self.registry.set(sess)
                return sess
        else:
            return self.registry()

    def remove(self):
        """Dispose of the current :class:`.Session`, if present.

        This will first call :meth:`.Session.close` method
        on the current :class:`.Session`, which releases any existing
        transactional/connection resources still being held; transactions
        specifically are rolled back.  The :class:`.Session` is then
        discarded.   Upon next usage within the same scope,
        the :class:`.scoped_session` will produce a new
        :class:`.Session` object.

        """

        if self.registry.has():
            self.registry().close()
        self.registry.clear()

    def configure(self, **kwargs):
        """reconfigure the :class:`.sessionmaker` used by this
        :class:`.scoped_session`.

        See :meth:`.sessionmaker.configure`.

        """

        if self.registry.has():
            self._logger.warning(
                "At least one scoped session is already present. "
                " configure() can not affect sessions that have "
                "already been created."
            )

        self.session_factory.configure(**kwargs)


default_session = scoped_session(session_maker())()



# *- coding: utf-8 -*-
"""
Foreign key component

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 28/01/2019
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from .decorators import classproperty
from .protocols import with_metaclass, SchemaMetaclass, ObjectProtocol
from .types import String
from .types.foreign_key import ForeignKey as ForeignKey_t


class Relationship(with_metaclass(SchemaMetaclass, ForeignKey_t)):
    """
    Class to deal with relationships
    """
    _id = 'https://numengo.org/ngoschema#/$defs/relationships/$defs/Relationship'
    _inheritance = False
    _reverse = False

    def __new__(cls, *args, **kwargs):
        new = super(ObjectProtocol, cls).__new__
        if new is object.__new__:
            return new(cls)
        return new(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        ObjectProtocol.__init__(self, *args, **kwargs)
        ForeignKey_t.__init__(self, *args, **kwargs)

    def set_foreignSchema(self, value):
        return ForeignKey_t.set_foreignSchema(self, value)

    def set_foreignClass(self, value):
        return ForeignKey_t.set_foreignClass(self, value)

    #def resolve(self, keys):
    #    return self._foreignClass.resolve_by_keys(keys)


class ForeignKey(with_metaclass(SchemaMetaclass)):
    """
    Class to deal with relationships
    """
    _id = 'https://numengo.org/ngoschema#/$defs/relationships/$defs/ForeignKey'

    def __new__(cls, *args, **kwargs):
        # to avoid to go through ObjectProtocol.__new__ which uses $schema for subclassing
        new = super(ObjectProtocol, cls).__new__
        if new is object.__new__:
            return new(cls)
        return new(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        Relationship.__init__(self, *args, **kwargs)

    @staticmethod
    def _serialize(self, value, **opts):
        if String.check(value):
            return {'foreignSchema': value}
        return ObjectProtocol._serialize(self, value, **opts)




import pytest

from future.utils import with_metaclass
from ngoschema.exceptions import InvalidValue

from ngoschema.managers import NamespaceManager


def test_namespace_manager():
    ns_mgr = NamespaceManager()
    uri = 'https://numengo.org/ngoschema/document#/$defs/Document'
    assert ns_mgr.get_cname_id('ngoschema.document.Document') == uri
    assert ns_mgr.get_id_cname(uri) == 'ngoschema.document.Document', ns_mgr.get_id_cname(uri)
    ns_mgr2 = NamespaceManager(document='https://numengo.org/ngoschema/document')
    assert ns_mgr2.get_id_cname(uri) == 'document.Document'
    assert ns_mgr2.load('document.Document')
    return


if __name__ == '__main__':
    test_namespace_manager()



content = """
  <body>
    <p id="onlinecourse">
      Long remained secret and restricted to circles of initiates, the 
      techniques of this <b>tantric ritual</b>&#160;are now available to you thanks 
      to our easy-to-follow course which includes:
    </p>
    <ul>
      <li>
        More than 130 slides with pictures and videos (<b>with explicit sexual 
        content</b>).
      </li>
      <li>
        Diagrams of the feminine anatomy to easily identify and locate all the 
        parts we will get to massage, such as her clitoris, her G-spot, 
        A-spot, K-spot, her cervix, etc...<br><a class="btn btn-primary span5" href="/yoni-massage/some-anatomy-first/" type="button">Learn more on feminine anatomy</a></br>
      </li>
      <li>
        Step-by-step instructions to create an amazing massage experience from 
        start to finish: learn how to give her a sensual full-body massage, 
        and then an incredible Yoni massage
      </li>
      <li>
        Clear descriptions of the different types of orgasms and how to give 
        them.<br><a class="btn btn-primary span5" href="/yoni-massage/some-anatomy-first/orgasms/" type="button">Learn more on orgasms</a></br>
      </li>
      <li>
        A brief introduction to Tantra. Understand the basic masculine and 
        feminine principles, chakras, and the role of sexual energy in this 
        spiritual practice.<br><a class="btn btn-primary span5" href="/about-tantra/">Learn more about Tantra</a></br>
      </li>
    </ul>
  </body>
"""

from ngoschema.utils import xmltodict
from pprint import pprint

html = xmltodict.parse(content)
body = html['body']
pprint(body)
print(xmltodict.unparse(html))



# -*- coding: utf-8 -*-
""" Unit tests for inspection utilities

test_inspect.py - created on 22/05/2018
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3  """

from __future__ import print_function
from __future__ import unicode_literals



def function_example(a, b=1, *args):
    """
    My dummy function
    :param a: string
    :return: integer
    """
    return b


def test_function():
    from ngoinsp.inspectors import inspect_class, inspect_function
    fi = inspect_function(function_example)
    assert fi['arguments'][0]['name'] == 'a'
    assert fi['arguments'][1]['name'] == 'b'
    assert fi['arguments'][0]['description']
    assert fi['returns']
    assert len(fi['arguments']) == 2
    assert fi['varargs']['name'] == 'args'
    assert not fi['keywords']
    assert not fi['decorators']


class Foo:
    a = 1

    def bar(self, b=1):
        return self.a + b


class Foo2(Foo):
    b = 1

    @staticmethod
    def bar_static(b):
        return b

    @classmethod
    def bar_cls(cls):
        return cls.b


def test_class():
    from ngoinsp.inspectors import inspect_class, inspect_function
    ci = inspect_class(Foo)
    assert ci['methods']['bar']
    ci = inspect_class(Foo2)
    assert ci['methods']['bar_static']['is_staticmethod']
    assert ci['methods']['bar_cls']['is_classmethod']
    assert 'bar' in ci['methods_inherited']


if __name__ == "__main__":
    test_function()
    test_class()




schema = {
    'type': 'object',
    'readOnly': ['canonicalName'],
    'properties': {
        'name': {
            'type': 'string',
            'default': '<anonymous>'
        },
        'canonicalName': {
            'type': 'string',
            'default': '{% if this.parent %}{{ this.parent.canonicalName }}.{% endif %}{{this.name}}',
        },
        'parent': {'$ref': '#'}
    }
}

from future.utils import with_metaclass
from ngoschema import get_builder, SchemaMetaclass, ProtocolBase


NamedClass = get_builder().construct('#', schema, (ProtocolBase, ))

a = NamedClass()
assert a.name == '<anonymous>'
a.name = 'a'
assert a.name == 'a', a.name
b = NamedClass(name='b')
assert b.canonicalName == 'b', b.canonicalName
b.parent = a
assert b.canonicalName == 'a.b', b.canonicalName

a.name = 'A'
assert a.name == 'A'
assert b.canonicalName == 'A.b', b.canonicalName

b.parent.name = 'a'
assert a.name == 'a'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

import io
import os
import re
import sys
import subprocess
from os.path import basename
from os.path import dirname
from os.path import join
from glob import glob
from os.path import splitext

import sys
sys.path.append(os.path.abspath('.'))


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')).read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    dir_path = dirname(os.path.realpath(__file__))
    bumpversion = open(join(dir_path, '.bumpversion.cfg')).read()
    return re.search("^current_version\s*=\s*(\S*)\s*\n",
                     bumpversion, re.MULTILINE).group(1)


name = 'ngoschema'
package = 'ngoschema'
description = 'automatic class-based binding to JSON schemas and toolkit for Domain-Driven Design.'
url = 'https://github.com/numengo/python-ngoschema'
author = 'Cédric ROMAN'
author_email = 'roman@numengo.com'
license = 'GNU General Public License v3'
version = get_version(package)


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(os.path.join(package))
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]
    filepaths = []
    for base, filenames in walk:
        filepaths.extend(
            [os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


setup_requires = [
    'matrix',
]

install_requires = [
    'pathlib',
    'future',
    'click',
    'jsonschema>=4.3',
    #'rfc3987',
    'ngofile',
    'attrs',
    'dpath',
    'pyrsistent',
    'simple_settings',
    'simple-settings[yaml,redis]',
    'appdirs',
    'wrapt',
    'jinja2',
    'arrow',
    'inflection',
    'slugify',
    'jsonpickle',
    'six',
    'requests',
    'ruamel.yaml',
    'python-magic-bin',
    'xmldict',
    'redis',
    'sqlalchemy',
]

post_install_requires = [i for i in install_requires if ('-' in i or ':' in i or '.' in i)]
install_requires = [i for i in install_requires if not ('-' in i or ':' in i or '.' in i)]


# for setuptools to work properly, we need to install packages with - or : separately
# and for that we need a hook
# https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools
class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        if post_install_requires:
            cmd = ['pip', 'install'] + post_install_requires
            print(cmd)
            subprocess.check_call(cmd)
        install.run(self)


test_requires = [
    'pytest',
    'pytest-logger',
]

extras_requires = {}

setup(
    name=name,
    version=version,
    license=license,
    description=description,
    long_description='%s\n%s' %
    (re.compile('^.. skip-next.*', re.M | re.S).sub('',
     re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('',
     read('README.rst'))),
     re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))),
    long_description_content_type='text/x-rst',
    author=author,
    author_email=author_email,
    url=url,
    packages=[package],
    package_data=get_package_data(package),
    include_package_data=True,
    zip_safe=False,
    keywords=[
        "json-schema", " schema", " class_builder", " data_validation",
        " type_checking", " mixins", " object_serialization",
        " code_generation"
    ],
    setup_requires=setup_requires,
    install_requires=install_requires,
    requires=install_requires,
    tests_require=test_requires,
    extras_require=extras_requires,
    entry_points={
        'console_scripts': [
        ]
    },
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    cmdclass={
        'install': PostInstallCommand,
        #'develop': PostInstallCommand, # not working with no-deps
    },
)

if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep wheel"):
        print("wheel not installed.\nUse `pip install wheel`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print("  git tag -a {0} -m 'version {0}'".format(version))
    print("  git push --tags")
    sys.exit()
